from contextlib import AbstractAsyncContextManager
from functools import partial
from typing import Any, AsyncIterator, Dict, Iterator, Optional

import aiohttp
import requests
from pydantic import StrictStr

from root.generated.openapi_client.api_client import ApiClient

from .generated.openapi_aclient import ApiClient as AApiClient
from .generated.openapi_aclient.api.v1_api import V1Api as ADatasetsApi
from .generated.openapi_aclient.models.data_set_create import DataSetCreate as ADataSetCreate
from .generated.openapi_aclient.models.data_set_list import DataSetList as ADataSetList
from .generated.openapi_aclient.models.paginated_data_set_list_list import (
    PaginatedDataSetListList as APaginatedDataSetListList,
)
from .generated.openapi_client.api.v1_api import V1Api as DatasetsApi
from .generated.openapi_client.models.data_set_create import DataSetCreate
from .generated.openapi_client.models.data_set_list import DataSetList
from .utils import ClientContextCallable, iterate_cursor_list, with_async_client, with_sync_client


class DataSets:
    """
    DataSets (sub) API

    Note:

      The construction of the API instance should be handled by
      accesing an attribute of a :class:`root.client.RootSignals` instance.
    """

    def __init__(self, client_context: ClientContextCallable, base_url: str, api_key: str):
        self.client_context = client_context
        self.base_url = base_url
        self.api_key = api_key

    def create(
        self,
        *,
        name: Optional[str] = None,
        path: Optional[str] = None,
        type: str = "reference",
        _request_timeout: Optional[int] = None,
    ) -> Optional[DataSetCreate]:
        """
        Create a dataset object with the given parameters to the registry.
        If the dataset has a path, it will be uploaded to the registry.
        """

        payload: Dict[str, Any] = {"name": name, "type": type, "tags": []}
        file = None
        try:
            if path:
                file = open(path, "rb")
                files = {"file": file}
            else:
                files = None

            response = requests.post(
                f"{self.base_url}/datasets/",
                headers={"Authorization": f"Api-Key {self.api_key}"},
                data=payload,
                files=files,
                timeout=_request_timeout or 120,
            )
            if not response.ok:
                raise Exception(f"create failed with status code {response.status_code} and message\n{response.text}")

            return DataSetCreate.from_dict(response.json())
        finally:
            if file and not file.closed:
                file.close()

    async def acreate(
        self,
        *,
        name: Optional[str] = None,
        path: Optional[str] = None,
        type: str = "reference",
        _request_timeout: Optional[int] = None,
    ) -> Optional[ADataSetCreate]:
        """
        Asynchronously create a dataset object with the given parameters to the registry.
        If the dataset has a path, it will be uploaded to the registry.

        """

        payload = aiohttp.FormData()
        payload.add_field("name", name)
        payload.add_field("type", type)

        file = None
        try:
            if path:
                file = open(path, "rb")
                payload.add_field("file", file)

            async with aiohttp.ClientSession() as session:
                async with session.post(
                    f"{self.base_url}/datasets/",
                    data=payload,
                    headers={"Authorization": f"Api-Key {self.api_key}"},
                    timeout=aiohttp.ClientTimeout(_request_timeout) or aiohttp.ClientTimeout(120),
                ) as response:
                    if not response.ok:
                        raise Exception(
                            f"create failed with status code {response.status} and message\n{response.text}"
                        )
                    return ADataSetCreate.from_dict(await response.json())
        finally:
            if file and not file.closed:
                file.close()

    @with_sync_client
    def get(
        self,
        dataset_id: str,
        *,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> DataSetList:
        """
        Get a dataset object from the registry.
        """

        api_instance = DatasetsApi(_client)
        return api_instance.v1_datasets_retrieve(dataset_id, _request_timeout=_request_timeout)

    @with_async_client
    async def aget(
        self,
        dataset_id: str,
        *,
        _request_timeout: Optional[int] = None,
        _client: AApiClient,
    ) -> ADataSetList:
        """
        Asynchronously get a dataset object from the registry.
        """

        api_instance = ADatasetsApi(_client)
        return await api_instance.v1_datasets_retrieve(dataset_id, _request_timeout=_request_timeout)

    @with_sync_client
    def list(
        self,
        search_term: Optional[str] = None,
        *,
        limit: int = 100,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> Iterator[DataSetList]:
        """
        Iterate through the datasets.

        Args:
          limit: Number of entries to iterate through at most.
          search_term: Can be used to limit returned datasets.
        """

        api_instance = DatasetsApi(_client)
        yield from iterate_cursor_list(
            partial(api_instance.v1_datasets_list, search=search_term, _request_timeout=_request_timeout), limit=limit
        )

    async def alist(
        self,
        search_term: Optional[str] = None,
        *,
        limit: int = 100,
        _request_timeout: Optional[int] = None,
    ) -> AsyncIterator[ADataSetList]:
        """
        Asynchronously iterate through the datasets.

        Args:
          limit: Number of entries to iterate through at most.
          search_term: Can be used to limit returned datasets.
        """

        context = self.client_context()
        assert isinstance(context, AbstractAsyncContextManager), "This method is not available in synchronous mode"
        async with context as client:
            api_instance = ADatasetsApi(client)
            partial_list = partial(api_instance.v1_datasets_list, search=search_term, _request_timeout=_request_timeout)
            cursor: Optional[StrictStr] = None
            while limit > 0:
                result: APaginatedDataSetListList = await partial_list(page_size=limit, cursor=cursor)
                if not result.results:
                    return

                used_results = result.results[:limit]
                limit -= len(used_results)

                for used_result in used_results:
                    yield used_result

                if not (cursor := result.next):
                    return

    @with_sync_client
    def delete(
        self,
        dataset_id: str,
        *,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> None:
        """
        Delete a dataset object from the registry.
        """

        api_instance = DatasetsApi(_client)
        return api_instance.v1_datasets_destroy(dataset_id, _request_timeout=_request_timeout)

    @with_async_client
    async def adelete(
        self,
        dataset_id: str,
        *,
        _request_timeout: Optional[int] = None,
        _client: AApiClient,
    ) -> None:
        """
        Asynchronously delete a dataset object from the registry.
        """

        api_instance = ADatasetsApi(_client)
        return await api_instance.v1_datasets_destroy(dataset_id, _request_timeout=_request_timeout)



from __future__ import annotations

from contextlib import AbstractAsyncContextManager
from functools import partial
from typing import TYPE_CHECKING, AsyncIterator, Iterator, List, Optional, Protocol

from pydantic import StrictStr

from .generated.openapi_aclient import ApiClient as AApiClient
from .generated.openapi_aclient.api.v1_api import V1Api as AExecutionLogsApi
from .generated.openapi_aclient.models.execution_log_details import ExecutionLogDetails as AExecutionLogDetails
from .generated.openapi_aclient.models.execution_log_list import ExecutionLogList as AExecutionLogList
from .generated.openapi_aclient.models.paginated_execution_log_list_list import (
    PaginatedExecutionLogListList as APaginatedExecutionLogListList,
)
from .generated.openapi_client import ApiClient
from .generated.openapi_client.api.v1_api import V1Api as ExecutionLogsApi
from .generated.openapi_client.models.execution_log_details import ExecutionLogDetails
from .generated.openapi_client.models.execution_log_list import ExecutionLogList
from .utils import ClientContextCallable, iterate_cursor_list, with_async_client, with_sync_client

if TYPE_CHECKING:

    class ExecutionResult(Protocol):
        execution_log_id: str


class ExecutionLogs:
    """Execution logs API"""

    def __init__(self, client_context: ClientContextCallable):
        self.client_context = client_context

    @with_sync_client
    def list(
        self,
        *,
        limit: int = 100,
        search_term: Optional[str] = None,
        tags: Optional[List[str]] = None,
        include: Optional[List[str]] = None,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> Iterator[ExecutionLogList]:
        """
        List execution logs

        Args:
          limit: Number of entries to iterate through at most.
          search_term: Can be used to limit returned logs. For example, a skill id or name.
          tags: Optional tags to filter the logs by.
          include: Optional fields to include in the response.
        """
        api_instance = ExecutionLogsApi(_client)
        yield from iterate_cursor_list(
            partial(
                api_instance.v1_execution_logs_list,
                search=search_term,
                tags=",".join(tags) if tags else None,
                include=",".join(include) if include else None,
                _request_timeout=_request_timeout,
            ),
            limit=limit,
        )

    async def alist(
        self,
        *,
        limit: int = 100,
        search_term: Optional[str] = None,
        tags: Optional[List[str]] = None,
        include: Optional[List[str]] = None,
        _request_timeout: Optional[int] = None,
    ) -> AsyncIterator[AExecutionLogList]:
        """
        Asynchronously list execution logs

        Args:
          limit: Number of entries to iterate through at most.
          search_term: Can be used to limit returned logs. For example, a skill id or name.
          tags: Optional tags to filter the logs by.
          include: Optional fields to include in the response.
        """

        context = self.client_context()
        assert isinstance(context, AbstractAsyncContextManager), "This method is not available in synchronous mode"
        async with context as client:
            api_instance = AExecutionLogsApi(client)
            partial_list = partial(
                api_instance.v1_execution_logs_list,
                search=search_term,
                tags=",".join(tags) if tags else None,
                include=",".join(include) if include else None,
                _request_timeout=_request_timeout,
            )

            cursor: Optional[StrictStr] = None
            while limit > 0:
                result: APaginatedExecutionLogListList = await partial_list(page_size=limit, cursor=cursor)
                if not result.results:
                    return

                used_results = result.results[:limit]
                limit -= len(used_results)

                for used_result in used_results:
                    yield used_result

                if not (cursor := result.next):
                    return

    @with_sync_client
    def get(
        self,
        *,
        log_id: Optional[str] = None,
        execution_result: Optional[ExecutionResult] = None,
        _request_timeout: Optional[int] = None,
        _client: ApiClient,
    ) -> ExecutionLogDetails:
        """
        Get a specific execution log details

        Args:
          log_id: The log to be fetched
          execution_result: The execution result containing the log ID.

        Raises:
          ValueError: If both log_id and execution_result are None.
        """

        api_instance = ExecutionLogsApi(_client)
        _log_id = execution_result.execution_log_id if execution_result else log_id
        if _log_id is None:
            raise ValueError("Either log_id or execution_result must be provided")
        return api_instance.v1_execution_logs_retrieve(_log_id, _request_timeout=_request_timeout)

    @with_async_client
    async def aget(
        self,
        *,
        log_id: Optional[str] = None,
        execution_result: Optional[ExecutionResult] = None,
        _request_timeout: Optional[int] = None,
        _client: AApiClient,
    ) -> AExecutionLogDetails:
        """
        Asynchronously get a specific execution log details

        Args:
          log_id: The log to be fetched
          execution_result: The execution result containing the log ID.

        Raises:
          ValueError: If both log_id and execution_result are None.
        """

        api_instance = AExecutionLogsApi(_client)
        _log_id = execution_result.execution_log_id if execution_result else log_id
        if _log_id is None:
            raise ValueError("Either log_id or execution_result must be provided")
        return await api_instance.v1_execution_logs_retrieve(_log_id, _request_timeout=_request_timeout)



from root import RootSignals

# Connect to the Root Signals API
client = RootSignals()

request = "Is the number of pensioners working more than 100k in 2023?"
response = "Yes, 150000 pensioners were working in 2024."

# Chunks retreived from a RAG pipeline
retreived_document_1 = """
While the work undertaken by seniors is often irregular and part-time, more than 150,000 pensioners were employed in 2023, the centre's statistics reveal. The centre noted that pensioners have increasingly continued to work for some time now.
"""
retreived_document_2 = """
According to the pension centre's latest data, a total of around 1.3 million people in Finland were receiving old-age pensions, with average monthly payments of 1,948 euros.
"""

# Measures is the answer faithful to my contexts (knowledge-base/documents)
faithfulness_result = client.evaluators.Faithfulness(
    request=request,
    response=response,
    contexts=[retreived_document_1, retreived_document_2],
)

print(faithfulness_result.score)  # 0.0 as the response does not match the retrieved documents
print(faithfulness_result.justification)

# Measures whether the retrieved context provides
# sufficient information to produce the ground truth response
context_recall_result = client.evaluators.Context_Recall(
    request="Was the number of pensioners who are working above 100k in 2023?",
    contexts=[retreived_document_1, retreived_document_2],
    expected_output="In 2023, 150k pensioners were still working.",  # Ground truth
)
print(context_recall_result.score)  # We expect a high score



from root import RootSignals
from root.validators import Validator

# Connect to the Root Signals API
client = RootSignals()

skill = client.skills.create(
    name="My strict chatbot",
    intent="Simple Q&A chatbot",
    prompt="Provide a clear answer to the question: {{question}}",
    model="gpt-4",
    validators=[Validator(evaluator_name="Clarity", threshold=0.6)],
)

# Execute the skill
response = skill.run({"question": "What is the capital of France?"})

# Get the validation results
print(response.validation)

# Get the execution details
log = client.execution_logs.get(execution_result=response)
print(log)

# List all execution logs
iterator = client.execution_logs.list(limit=10)
print(next(iterator))



from root import RootSignals

# Connect to the Root Signals API
client = RootSignals()

result = client.evaluators.Helpfulness(response="You can find the instructions from our Careers page.")

print(f"Score: {result.score} / 1.0")  # A normalized score between 0 and 1
print(result.justification)  # The reasoning for the score



from root import RootSignals
from root.data_loader import WebPageDataLoader

# Create a client
client = RootSignals()

# Create a skill
skill = client.skills.create(
    name="My kimchi cooking assistant",
    intent="This skill will find you a recipe for kimchi.",
    prompt=("Help me prepare Kimchi. Use this web page for ingredients {{kimchi_ingredients}}"),
    model="gpt-4o",
    fallback_models=["gpt-4"],
    data_loaders=[
        WebPageDataLoader(
            name="kimchi_ingredients",
            url="https://simple.wikipedia.org/wiki/Kimchi",
        )
    ],
)
response = skill.run()
print(response)
